#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <errno.h>
#include <string.h>
#include <mthread/mthread_internal.h>

/* Functions for mutex handling.  */

/* Initialize MUTEX using attributes in *MUTEX_ATTR, or use the
   default values if later is NULL.  */
int
mthread_mutex_init(mthread_mutex_t *__mutex, const mthread_mutexattr_t *__mutex_attr)
{
    UNUSED(__mutex_attr);
    __mutex->list        = safe_malloc(sizeof(mthread_list_t));
    __mutex->list->first = NULL;
    __mutex->list->last  = NULL;
    __mutex->nb_thread   = 0;
    __mutex->lock        = 0;
    LOG_INFO("MTHREAD", "%s", "MUTEX initialized");
    return 0;
}

/* Destroy MUTEX.  */
int
mthread_mutex_destroy(mthread_mutex_t *__mutex)
{
    mthread_spinlock_lock(&__mutex->lock);

    if (__mutex->nb_thread != 0)
        return EBUSY;
    free(__mutex->list);
    __mutex->list = NULL;

    mthread_spinlock_unlock(&__mutex->lock);
    return 0;
}

/* Try to lock MUTEX.  */
int
mthread_mutex_trylock(mthread_mutex_t *__mutex)
{
    int retval = EINVAL;

    if (__mutex->list == NULL) {
        __mutex->list = safe_malloc(sizeof(mthread_list_t));
        memset(__mutex->list, 0, sizeof(mthread_list_t));
    }

    mthread_spinlock_lock(&__mutex->lock);

    if (__mutex->nb_thread == 0) {
        __mutex->nb_thread = 1;
        mthread_spinlock_unlock(&__mutex->lock);
        retval = 0;
        goto end;
    }

    mthread_spinlock_unlock(&__mutex->lock);
    LOG_WARN("MTHREAD", "%s", "MUTEX busy");
    retval = EBUSY;

end:
    return retval;
}

/* Wait until lock for MUTEX becomes available and lock it.  */
int
mthread_mutex_lock(mthread_mutex_t *__mutex)
{
    mthread_t self;
    mthread_virtual_processor_t *vp;

    if (__mutex->list == NULL) {
        __mutex->list = safe_malloc(sizeof(mthread_list_t));
        memset(__mutex->list, 0, sizeof(mthread_list_t));
    }

    mthread_spinlock_lock(&__mutex->lock);

    if (__mutex->nb_thread == 0) {
        __mutex->nb_thread = 1;
        mthread_spinlock_unlock(&__mutex->lock);
    } else {
        self = mthread_self();
        mthread_insert_last(self, __mutex->list);
        self->status = BLOCKED;
        vp = mthread_get_vp();
        vp->p = &__mutex->lock;
        mthread_yield();
    }

    return 0;
}

/* Unlock MUTEX.  */
int
mthread_mutex_unlock(mthread_mutex_t *__mutex)
{
    mthread_t first;
    mthread_virtual_processor_t *vp;

    if (__mutex->list == NULL) {
        __mutex->list = safe_malloc(sizeof(mthread_list_t));
        memset(__mutex->list, 0, sizeof(mthread_list_t));
    }

    mthread_spinlock_lock(&__mutex->lock);
    if (__mutex->list->first != NULL) {
        first = mthread_remove_first(__mutex->list);
        vp = mthread_get_vp();
        first->status = RUNNING;
        mthread_insert_last(first, &(vp->ready_list));
    } else
        __mutex->nb_thread = 0;

    mthread_spinlock_unlock(&__mutex->lock);
    return 0;
}
