#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <common/queue.h>
#include <lektor/commands.h>
#include <lektor/database.h>
#include <lektor/net.h>
#include <lektor/reg.h>

#include <sqlite3.h>

#include <strings.h>
#include <sched.h>
#include <assert.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <limits.h>

enum lkt_command_list_mode {
    LKT_COMMAND_LIST_OFF,
    LKT_COMMAND_LIST_ON,
    LKT_COMMAND_LIST_ON_OK,
};

struct lkt_client {
    char buffer_in[LKT_MESSAGE_MAX + 1];
    size_t buffer_in_len;

    struct lkt_message *buffer_out[BUFFER_OUT_MAX];
    size_t buffer_out_len;

    struct lkt_command command_list[COMMAND_LIST_MAX];
    size_t command_list_len;
    enum lkt_command_list_mode command_list_mode;

    mpd_idle_flag mpd_idle_watch;
    bool request_close;

    bool authentificated;
    int continuation;
};

void
lkt_set_event_available(struct lkt_state *srv, enum lkt_event_type type)
{
    lkt_queue_make_available(&srv->queue, type);
}

static inline bool
lkt_close_client(struct lkt_state *srv, size_t c)
{
    return srv->clients[c - 1].request_close = true;
}

inline int
lkt_get_continuation(struct lkt_state *srv, size_t c)
{
    return srv->clients[c - 1].continuation;
}

inline void
lkt_set_continuation(struct lkt_state *srv, size_t c, int i)
{
    srv->clients[c - 1].continuation = i;
}

inline size_t
lkt_remaining_msg(struct lkt_state *srv, size_t c)
{
    return BUFFER_OUT_MAX - srv->clients[c - 1].buffer_out_len;
}

void
lkt_state_send(struct lkt_state *srv, size_t c, struct lkt_message *msg)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    RETURN_IF(cli->buffer_out_len >= BUFFER_OUT_MAX, "Cannot send message to "
              "client, buffer_out is full", NOTHING);
    cli->buffer_out[cli->buffer_out_len++] = msg;
    srv->fds[c].events |= POLLOUT;
}

static inline void
send_continue(struct lkt_state *srv, size_t c, int i)
{
    struct lkt_message *cont = lkt_message_new();
    cont->data_len = safe_snprintf(cont->data, LKT_MESSAGE_MAX,
                                   "continue: %d\n", i);
    lkt_state_send(srv, c, cont);
}

static void
send_ok(struct lkt_state *srv, size_t c)
{
    struct lkt_message *ok = lkt_message_new();
    if (srv->clients[c - 1].command_list_mode == LKT_COMMAND_LIST_ON_OK) {
        strncpy(ok->data, "list_OK\n", 9);
        ok->data_len = 8;
    } else {
        strncpy(ok->data, "OK\n", 4);
        ok->data_len = 3;
    }
    lkt_state_send(srv, c, ok);
}

static void
send_ack(struct lkt_state *srv, size_t c, const char *command)
{
    struct lkt_message *ack = lkt_message_new();
    ack->data_len = snprintf(ack->data, LKT_MESSAGE_MAX,
                             "ACK [1@0] {%s} failed\n", command);
    lkt_state_send(srv, c, ack);
}

static void
send_status(struct lkt_state *srv, size_t c, int status, const char *cmd_name)
{
    if (!status) {
        LOG_INFO("COMMAND", "Command '%s'", cmd_name);
        send_ok(srv, c);
    } else {
        LOG_INFO("COMMAND", "Command '%s' %s", cmd_name,
                 status == 2 ? "is unknown" : "has failed");
        send_ack(srv, c, cmd_name);
    }
}

static int
command_list_begin(struct lkt_state *srv, size_t c, int list_ok)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    if (cli->command_list_mode != LKT_COMMAND_LIST_OFF)
        return -1;

    if (list_ok)
        cli->command_list_mode = LKT_COMMAND_LIST_ON_OK;
    else
        cli->command_list_mode = LKT_COMMAND_LIST_ON;

    return 0;
}

static int
handle_simple_command(struct lkt_state *srv, size_t c, struct lkt_command cmd)
{
    int err = 0, continuation = 0;

    switch (lkt_client_get_mask(srv, c)) {
    case MPD_IDLE_NONE:
        /* Commands that requires authentification. */
        if (STR_MATCH(cmd.name, "__adduser"))
            err = ! command_user_add(srv, c, srv->db, cmd.args);
        else if (STR_MATCH(cmd.name, "__restart"))
            err = ! command_restart(srv, c);
        else if (STR_MATCH(cmd.name, "kill"))
            err = ! command_kill(srv, c);
        else if (STR_MATCH(cmd.name, "update"))
            err = ! command_update(srv, c, cmd.args);
        else if (STR_MATCH(cmd.name, "rescan"))
            err = ! command_rescan(srv, c, cmd.args, false);
        else if (STR_MATCH(cmd.name, "__rescan"))
            err = ! command_rescan(srv, c, cmd.args, true);

        /* Commands that are available if not in idle mode */
        else if (STR_MATCH(cmd.name, "currentsong"))
            err = !command_currentsong(srv, c);
        else if (STR_MATCH(cmd.name, "status"))
            err = !command_status(srv, c);

        else if (STR_MATCH(cmd.name, "close"))
            err = !lkt_close_client(srv, c);
        else if (STR_MATCH(cmd.name, "ping"))
            err = 0;

        else if (STR_MATCH(cmd.name, "next"))
            err = !command_next(srv->db, &srv->win, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "pause"))
            err = !command_pause(srv->db, &srv->win, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "previous"))
            err = !command_previous(srv->db, &srv->win, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "play"))
            err = ! command_play(srv->db, &srv->win, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "playid"))
            err = ! command_playid(srv->db, &srv->win, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "stop"))
            err = !command_stop(srv->db, &srv->win, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "clear"))
            err = !command_clear(srv->db, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "crop"))
            err = !command_crop(srv->db, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "moveid"))
            err = !command_move(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "shuffle"))
            err = !command_shuffle(srv->db, &srv->mpd_idle_events);

        else if (STR_MATCH(cmd.name, "playlist") || STR_MATCH(cmd.name, "playlistinfo"))
            err = !command_queue_list(srv, c, cmd.args);
        else if (STR_MATCH(cmd.name, "playlistfind") || STR_MATCH(cmd.name, "playlistsearch"))
            err = ! command_find(srv, c, cmd.args, cmd.cont, database_search_queue_init);

        else if (STR_MATCH(cmd.name, "sticker") && cmd.args[0]) {
            if (STR_MATCH(cmd.args[0], "get") || STR_MATCH(cmd.args[0], "list") || STR_MATCH(cmd.args[0], "find"))
                err = ! command_sticker_get(srv, c, &cmd.args[1]);
            else if (STR_MATCH(cmd.args[0], "set"))
                err = ! command_sticker_set(srv, c, &cmd.args[1]);
            else if (STR_MATCH(cmd.args[0], "delete"))
                err = ! command_sticker_delete(srv, c, &cmd.args[1]);
            else if (STR_MATCH(cmd.args[0], "__create"))
                err = ! command_sticker_create(srv, c, &cmd.args[1]);
        }

        else if (STR_MATCH(cmd.name, "help"))
            err = !command_help(srv, c);

        else if (STR_MATCH(cmd.name, "__insert"))
            err = ! command_add(srv->db, &srv->win, cmd.args,
                                &srv->mpd_idle_events, 5);
        else if (STR_MATCH(cmd.name, "searchadd")   ||
                 STR_MATCH(cmd.name, "findadd")     ||
                 STR_MATCH(cmd.name, "add"))
            err = ! command_add(srv->db, &srv->win, cmd.args,
                                &srv->mpd_idle_events, 1);
        else if (STR_MATCH(cmd.name, "addid"))
            err = ! command_addid(srv->db, &srv->win, cmd.args,
                                  &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "deleteid"))
            err = ! (cmd.args[0] != NULL &&
                     command_delid(srv->db, &srv->win, cmd.args[0],
                                   &srv->mpd_idle_events));

        else if (STR_MATCH(cmd.name, "playlistclear"))
            err = ! command_plt_clear(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "rename"))
            err = ! command_plt_rename(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "playlistdelete"))
            err = ! command_plt_remove(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "playlistadd"))
            err = ! command_plt_add(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "rm") && cmd.args[0] != NULL && cmd.args[1] == NULL)
            err = ! command_plt_remove(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "save"))
            err = ! command_plt_export(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "__import"))
            err = ! command_plt_import(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "__dump"))
            err = ! command_dump(srv->db, cmd.args, &srv->mpd_idle_events);
        else if (STR_MATCH(cmd.name, "listplaylist"))
            err = ! command_find(srv, c, cmd.args, cmd.cont, database_search_playlist_init);

        else if (STR_MATCH(cmd.name, "random"))
            err = !command_set_playback_option(srv, c, lkt_playback_option_random, cmd.args);
        else if (STR_MATCH(cmd.name, "repeat"))
            err = !command_set_playback_option(srv, c, lkt_playback_option_repeat, cmd.args);
        else if (STR_MATCH(cmd.name, "setvol"))
            err = !command_set_playback_option(srv, c, lkt_playback_option_volume, cmd.args);
        else if (STR_MATCH(cmd.name, "single"))
            err = !command_set_playback_option(srv, c, lkt_playback_option_single, cmd.args);
        else if (STR_MATCH(cmd.name, "consume"))
            err = !command_set_playback_option(srv, c, lkt_playback_option_consume, cmd.args);

        else if (STR_MATCH(cmd.name, "password"))
            err = !command_password(srv, c, cmd.args);

        else if (STR_MATCH(cmd.name, "idle")) {
            err = !command_idle(srv, c, &cmd);
            goto end_no_send_status;
        } else if (STR_MATCH(cmd.name, "search") || STR_MATCH(cmd.name, "find"))
            err = ! command_find(srv, c, cmd.args, cmd.cont, database_search_database_init);
        else
            err = 2;

        break;

    default:
        /* commands available only in idle mode */
        if (STR_MATCH(cmd.name, "idle")) {
            err = !command_idle(srv, c, &cmd);
            goto end_no_send_status;
        } else if (STR_MATCH(cmd.name, "noidle"))
            err = !command_noidle(srv, c);
        else
            err = 2;

        break;
    }

    continuation = lkt_get_continuation(srv, c);
    if (continuation > 0) {
        LOG_INFO("NETWORK", "Client should continue from '%d'",
                 continuation);
        send_continue(srv, c, continuation);
    }
    send_status(srv, c, err, cmd.name);
end_no_send_status:
    return err;
}

static int
command_list_end(struct lkt_state *srv, size_t c)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    int err = 0;

    if (cli->command_list_mode == LKT_COMMAND_LIST_OFF)
        return -1;

    for (size_t i = 0; i < cli->command_list_len; i++) {
        err |= handle_simple_command(srv, c, cli->command_list[i]);
        lkt_command_free(&cli->command_list[i]);
    }

    cli->command_list_mode = LKT_COMMAND_LIST_OFF;
    cli->command_list_len = 0;

    if (!err)
        send_ok(srv, c);

    return err;
}

static int
handle_command(struct lkt_state *srv, size_t i, struct lkt_command cmd)
{
    struct lkt_client *cli = &srv->clients[i - 1];
    int err = 0;

    if (STR_MATCH(cmd.name, "command_list_begin"))
        err = command_list_begin(srv, i, 0);
    else if (STR_MATCH(cmd.name, "command_list_ok_begin"))
        err = command_list_begin(srv, i, 1);
    else if (STR_MATCH(cmd.name, "command_list_end"))
        err = command_list_end(srv, i);
    else if (cli->command_list_mode != LKT_COMMAND_LIST_OFF)
        cli->command_list[cli->command_list_len++] = lkt_command_clone(&cmd);
    else
        err = handle_simple_command(srv, i, cmd);

    LOG_INFO("COMMAND", "Command result '%s' -> %d", cmd.name, err);
    return err;
}

static int
handle_incoming_data(struct lkt_state *srv, size_t i)
{
    struct lkt_client *cli = &srv->clients[i - 1];
    int n;

    for (;;) {
        /* Recieve some data. */
        n = recv(srv->fds[i].fd, cli->buffer_in + cli->buffer_in_len,
                 LKT_MESSAGE_MAX - cli->buffer_in_len, 0);
        if (n < 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN)
                return 0;
            return -1;
        }
        if (n == 0)
            return -1;
        cli->buffer_in_len += n;

        /* Handle the data. */
        char *buf = cli->buffer_in;
        for (;;) {
            size_t line_len = strcspn(buf, "\n");
            if (line_len >= cli->buffer_in_len)
                break;
            buf[line_len] = 0;

            struct lkt_command cmd = lkt_command_parse(buf);
            handle_command(srv, i, cmd);

            buf += line_len + 1;
            cli->buffer_in_len -= line_len + 1;
        }

        if (cli->buffer_in < buf)
            memmove(cli->buffer_in, buf, cli->buffer_in_len);
        else if (cli->buffer_in_len == LKT_MESSAGE_MAX)
            return -1;
    }

    return -1;
}

static inline void
handle_disconnected_client(struct lkt_state *srv, size_t c)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    for (size_t i = 0; i < cli->buffer_out_len; i++)
        lkt_message_free(cli->buffer_out[i]);
    cli->buffer_out_len = 0;
}

static int
handle_outgoing_data(struct lkt_state *srv, size_t c)
{
    struct lkt_client *cli = &srv->clients[c - 1];

    for (size_t i = 0; i < cli->buffer_out_len; i++) {
        struct lkt_message *msg = cli->buffer_out[i];
        int n = send(srv->fds[c].fd, msg->data, msg->data_len, MSG_NOSIGNAL);
        if (n <= 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN) {
                cli->buffer_out_len -= i;
                memmove(cli->buffer_out, cli->buffer_out + i,
                        cli->buffer_out_len);
                return 0;
            }
            if (errno == EPIPE) {
                LOG_WARN("NETWORK", "Client %ld is out, free all its "
                         "messages", c);
                handle_disconnected_client(srv, i);
            }
            return -1;
        }

        lkt_message_free(msg);
    }

    cli->buffer_out_len = 0;
    srv->fds[c].events &= ~POLLOUT;
    return 0;
}

static int
init_listening_socket(const char *host, const char *port)
{
    const int yes = 1;
    int fd = -1;
    struct addrinfo *available;
    struct addrinfo hints = {0};

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;
    if (getaddrinfo(host, port, &hints, &available) < 0) {
        perror("getaddrinfo failed");
        return -1;
    }

    for (struct addrinfo *p = available; p; p = p->ai_next) {
        fd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
        if (fd < 0) {
            perror("Failed to create socket");
            continue;
        }

        if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0) {
            perror("Failed to reuse address");
            goto failure;
        }

        int flags = fcntl(fd, F_GETFL, 0);
        if (flags < 0) {
            perror("Failed to get socket flag");
            goto failure;
        }

        if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
            perror("Failed to set non-blocking flag on socket");
            goto failure;
        }

        if (bind(fd, p->ai_addr, p->ai_addrlen) < 0) {
            close(fd);
            perror("Failed to bind socket");
            continue;
        }

        break;

failure:
        close(fd);
        fd = -1;
        break;
    }

    freeaddrinfo(available);

    if (fd < 0) {
        LOG_ERROR("NETWORK", "Failed to bind to %s:%s",
                  host ? host : "0.0.0.0", port);
        return -1;
    }

    if (listen(fd, LKT_BACKLOG) < 0) {
        LOG_ERROR("NETWORK", "Failed to listen: %s", strerror(errno));
        return -1;
    }

    LOG_INFO("NETWORK", "Listening on %s:%s", host ? host : "0.0.0.0", port);
    return fd;
}

static int
accept_all(int listen_fd, struct pollfd *fds, size_t fds_max, size_t *fds_len)
{
    int fd;
    struct sockaddr_in peer_addr;
    socklen_t pa_size = sizeof(peer_addr);
    fds += *fds_len;

    for (int n = 0;; n++) {
        fd = accept(listen_fd, (struct sockaddr *) &peer_addr, &pa_size);
        if (fd < 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN)
                return n;
            if (errno == ECONNABORTED)
                continue;
            perror("accept");
            return -1;
        }

        if (*fds_len == fds_max) {
            LOG_ERROR("NETWORK", "Maximum number '%ld' of file descriptors"
                      " reached", fds_max);
            break;
        }

        int flag = fcntl(fd, F_GETFL, 0);
        if (flag < 0) {
            perror("fcntl");
            break;
        }
        if (fcntl(fd, F_SETFL, flag | O_NONBLOCK) < 0) {
            perror("fcntl");
            break;
        }

        char *host = inet_ntoa(peer_addr.sin_addr);
        uint16_t port = htons(peer_addr.sin_port);
        LOG_WARN("NETWORK", "New connection from %s:%d", host, port);

        fds->fd = fd;
        fds->events = POLLIN;
        fds++;
        (*fds_len)++;
    }

    shutdown(fd, SHUT_RDWR);
    close(fd);
    return -1;
}

static int
handle_network_events(struct lkt_state *srv)
{
    int n = poll(srv->fds, srv->fds_len, 100);
    if (n < 0 && errno != EAGAIN && errno != EINTR) {
        perror("poll failed");
        return -1;
    }
    if (n < 0)
        return 0;

    if (srv->fds[0].revents & POLLIN) {
        int n = accept_all(srv->fds[0].fd, srv->fds,
                           srv->fds_max, &srv->fds_len);
        if (n < 0)
            return -1;
        struct lkt_message *msg = lkt_message_new();
        msg->data_len = snprintf(msg->data, LKT_MESSAGE_MAX,
                                 "OK MPD " MPD_VERSION "\n");
        for (; n; n--) {
            memset(&srv->clients[srv->fds_len - n - 1], 0,
                   sizeof(struct lkt_client));
            lkt_state_send(srv, srv->fds_len - n, lkt_message_clone(msg));
        }
        lkt_message_free(msg);
    }

    for (size_t i = 1; i < srv->fds_len; i++) {
        if (!srv->fds[i].revents)
            continue;
        if (srv->fds[i].revents & (POLLHUP | POLLERR)) {
            close(srv->fds[i].fd);
            srv->fds[i].fd = -1;
            srv->fds[i].revents &= ~ (POLLHUP | POLLERR);
            handle_disconnected_client(srv, i);
            LOG_WARN("NETWORK", "Client %ld is out", i);
            continue;
        }
        if (srv->fds[i].revents & POLLIN) {
            if (handle_incoming_data(srv, i) < 0) {
                close(srv->fds[i].fd);
                srv->fds[i].fd = -1;
            }
        }
        if (srv->fds[i].revents & POLLOUT) {
            if (handle_outgoing_data(srv, i) < 0) {
                close(srv->fds[i].fd);
                srv->fds[i].fd = -1;
            }
        }
        if (srv->clients[i - 1].request_close) {
            handle_outgoing_data(srv, i);
            close(srv->fds[i].fd);
            srv->fds[i].fd = -1;
        }
    }

    for (size_t i = 1; i < srv->fds_len; i++) {
        if (srv->fds[i].fd > 0)
            continue;

        LOG_WARN("NETWORK", "Connection closed by client %ld", i);
        srv->fds[i] = srv->fds[srv->fds_len - 1];
        srv->clients[i - 1] = srv->clients[srv->fds_len - 2];
        srv->fds_len--;
        i--;
    }

    return 0;
}

static int
handle_idle_events(struct lkt_state *srv)
{
    struct lkt_client *current_client;
    struct lkt_message *msg;
    mpd_idle_flag common_mask;
    bool once = false;

    // No new events
    if (srv->mpd_idle_events == MPD_IDLE_NONE)
        return 0;

    for (size_t i = 1; i < srv->fds_len; ++i) {
        current_client = &srv->clients[i - 1];

        // Client is not idle
        if (current_client->mpd_idle_watch == MPD_IDLE_NONE)
            continue;

        common_mask = current_client->mpd_idle_watch & srv->mpd_idle_events;

        if (MPD_IDLE_NONE == common_mask)
            continue;

        msg = lkt_message_new();
        memset(msg->data, 0, LKT_MESSAGE_MAX);
        snprintf(msg->data, LKT_MESSAGE_MAX, "changed: ");
        once |= true;

        if (MPD_IDLE_DATABASE & common_mask)
            strncat(msg->data, "database ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_UPDATE & common_mask)
            strncat(msg->data, "update ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_STORED_PLAYLIST & common_mask)
            strncat(msg->data, "stored_playlist ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PLAYLIST & common_mask)
            strncat(msg->data, "playlist ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PLAYER & common_mask)
            strncat(msg->data, "player ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_MIXER & common_mask)
            strncat(msg->data, "mixer ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_OUTPUT & common_mask)
            strncat(msg->data, "output ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_OPTIONS & common_mask)
            strncat(msg->data, "options ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PARTITION & common_mask)
            strncat(msg->data, "partition ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_STICKER & common_mask)
            strncat(msg->data, "sticker ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_SUBSCRIPTION & common_mask)
            strncat(msg->data, "subscritpion ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_MESSAGE & common_mask)
            strncat(msg->data, "message ", LKT_MESSAGE_MAX - 1);

        msg->data_len = strlen(msg->data);
        LOG_INFO("COMMAND", "Sending '%s' (len: %ld) to client %ld",
                 msg->data, msg->data_len, i);
        msg->data[msg->data_len - 1] = '\n';
        msg->data[msg->data_len] = '\0';
        lkt_state_send(srv, i, msg);
        send_ok(srv, i);
    }

    if (once)
        srv->mpd_idle_events = MPD_IDLE_NONE;

    return 0;
}

int
lkt_client_get_mask(struct lkt_state *srv, size_t c)
{
    return srv->clients[c - 1].mpd_idle_watch;
}

void
lkt_client_add_mask(struct lkt_state *srv, size_t c, mpd_idle_flag add)
{
    if (pthread_mutex_lock(&srv->lock))
        return;
    srv->clients[c - 1].mpd_idle_watch |= add;
    pthread_mutex_unlock(&srv->lock);
}

void
lkt_client_clear_mask(struct lkt_state *srv, size_t c)
{
    if (pthread_mutex_lock(&srv->lock))
        return;
    srv->clients[c - 1].mpd_idle_watch = MPD_IDLE_NONE;
    pthread_mutex_unlock(&srv->lock);
}

inline bool
lkt_client_auth(struct lkt_state *srv, size_t c, bool set)
{
    return srv->clients[c - 1].authentificated |= set;
}

static inline void
handle_queue_events(struct lkt_state *srv)
{
    lkt_event evt;
    char *string = malloc(BUFFER_MAX * sizeof(char));
    if (!string)
        return;
redo:
    evt = lkt_queue_handle(&srv->queue);

    switch (evt.type) {
    case lkt_event_play_pos:
        safe_snprintf(string, BUFFER_MAX, "%ld", (size_t) evt.attr);
        command_play(srv->db, &srv->win, &string, &srv->mpd_idle_events);
        break;

    case lkt_event_null:
    default:
        free(string);
        return;
    }
    goto redo;
}

void
lkt_listen(struct lkt_state *srv)
{
    srv->fds_max = 16;
    srv->fds     = calloc(srv->fds_max, sizeof(struct pollfd));
    srv->clients = calloc(srv->fds_max, sizeof(struct lkt_client));
    memset(srv->clients, 0, srv->fds_max * sizeof(struct lkt_client));

    if ((srv->fds[0].fd = init_listening_socket(srv->host, srv->port)) < 0)
        return;

    srv->fds[0].events = POLLIN;
    srv->fds_len = 1;

    /* Listen */
    for (;;) {
        if (handle_network_events(srv) < 0)
            break;
        if (handle_idle_events(srv) < 0)
            break;
        srv->win.handle_events(&srv->win, srv->db, &srv->mpd_idle_events);
        handle_queue_events(srv);
    }
}
